#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 14.01.2020


from    .traverse2d     import  traverse2d
from    .traverse3d     import  traverse3d

__all__ = [
    "traverse2d",
    "traverse3d",
]


__license__ =       "GNU v3.0"
__maintainer__ =    "Andrei Leonard Nicusan"
__email__ =         "a.l.nicusan@bham.ac.uk"
__status__ =        "Beta"
