#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __version__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019


VERSION = (0, 3, 1)

__version__ = '.'.join(map(str, VERSION))
