#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


from    .modular_camera    import  ModularCamera


__all__ = [
    'ModularCamera'
]


__author__ = "Sam Manger"
__credits__ = [
    "Andrei Leonard Nicusan",
    "Kit Windows-Yule",
    "Sam Manger"
]
__license__ = "GNU v3.0"
__maintainer__ = "Sam Manger"
__email__ = "s.manger@bham.ac.uk"
__status__ = "Development"


